"""
change this file to your custom requirements (only import and perform as
little preprocessing as possible)
"""

from aixweather.imports.utils_import import MetaData


def load_custom_meta_data():
    """
    define or load your meta data here
    """

    meta = MetaData()
    meta.station_name = ""
    meta.input_source = f"Custom data"
    meta.altitude = ""
    meta.longitude = ""
    meta.latitude = ""

    return meta


def load_custom_from_file(path):
    """
    Import data from file and convert them into df

    Args:
        path: absolute path to file

    Returns:
        pd.Dataframe: as raw as possible but with some datetimeindex
    """

    ### load file to Dataframe

    ### convert timestamp to datetime and set as index

    ### return data frame
    return "DataFrame"



"""
AixWeather Imports Package

The 'imports' package within AixWeather contains modules for importing raw weather data from various sources,
including pulling data from DWD (Deutscher Wetterdienst) or reading files from local drives.
Each module is designed to handle data from a specific source or format.
"""


"""
includes auxiliary functions for data handling and transformation
"""

import pandas as pd
import numpy as np

def force_data_variable_convention(
    df: pd.DataFrame, format_desired: dict
) -> pd.DataFrame:
    """
    Ensure that all and only desired variable names are present and in correct order.

    Args:
        df (pd.DataFrame): The DataFrame containing the data to be formatted.
        format_desired (dict): A dictionary specifying the desired format with variable names as keys.

    Returns:
        pd.DataFrame: A DataFrame with filtered data in the desired format and order.
    """

    # filter existing df
    desired_var_names = set(format_desired.keys())
    df_core_var = df.loc[:, df.columns.isin(desired_var_names)]

    # Reindex the DataFrame to ensure all required columns exist and obey the order of columns
    df_core_var = df_core_var.reindex(columns=format_desired.keys())

    return df_core_var


def rename_columns(df: pd.DataFrame, format_dict: dict) -> pd.DataFrame:
    """
    Rename DataFrame columns based on the provided format dictionary.

    Args:
        df (pd.DataFrame): The DataFrame whose columns need to be renamed.
        format_dict (dict): A dictionary specifying the column renaming mapping,
                            with current column names as keys and desired names as values.

    Returns:
        pd.DataFrame: A DataFrame with renamed columns.
    """
    rename_map = {key: val["core_name"] for key, val in format_dict.items()}
    return df.rename(columns=rename_map)


def fill_nan_from_format_dict(df: pd.DataFrame, format_data: dict) -> pd.DataFrame:
    """
    Fill NaN values in a DataFrame based on the provided format data.

    Args:
        df (pd.DataFrame): The DataFrame containing the data to be processed.
        format_data (dict): A dictionary specifying NaN replacement values for columns,
                            with column names as keys and NaN replacement values as values.

    Returns:
        pd.DataFrame: A DataFrame with NaN values filled as per the format data.
    """
    nan_key = "nan"
    for key, value in format_data.items():
        nan = value[nan_key]
        if nan is not None:
            df[key].fillna(nan, inplace=True)
    return df


def replace_dummy_with_nan(df: pd.DataFrame, format_dict: dict) -> pd.DataFrame:
    """
    Replace specific values in the DataFrame with NaN based on the given format dictionary.
    Reason: sometimes, e.g. the DWD, specifies a missing value with a dummy value like e.g. 99,
    which makes it hard to see where missing values are and might affect the simulation.

    Args:
        df (pd.DataFrame): The DataFrame to be processed.
        format_dict (dict): A dictionary specifying values to be replaced with NaN,
                            with column names as keys and corresponding dummy values as values.

    Returns:
        pd.DataFrame: A DataFrame with specified values replaced by NaN.
    """

    for key, value in format_dict.items():
        if "nan" in value and key in df.columns:
            nan_values = value["nan"]
            if not isinstance(nan_values, list):
                nan_values = [nan_values]
            for nan_val in nan_values:
                df[key] = df[key].replace(nan_val, np.nan)
    return df


def evaluate_transformations(core_format: dict, other_format: dict):
    """
    Compare the units and core variables of two formats and print any required unit transformations.

    Args:
        core_format (dict): A dictionary representing the core format with keys as variable names and values
                            containing unit information.
        other_format (dict): A dictionary representing another format to be compared with the core format.
                            It contains keys and values with 'core_name' and 'unit' attributes.

    Raises:
        ValueError: If a core variable in other_format doesn't match the core variable format.
    """

    print("\nEvaluate format.")
    for key, value in other_format.items():
        if value["core_name"] in core_format.keys():
            # compare units
            if value["unit"] != core_format[value["core_name"]]["unit"]:
                print(
                    f"Unit transformation required for {value['core_name']} from"
                    f" {value['unit']} to {core_format[value['core_name']]['unit']}."
                )
        elif not value["core_name"]:
            pass
        else:
            raise ValueError(
                f"The core variable '{value['core_name']}' of import variable"
                f" {key} does not fit the core variable format"
            )


def select_entry_by_core_name(format_dict: dict, core_name_to_match: str):
    """
    Select an entry from a format dictionary based on the specified core name.

    Args:
        format_dict (dict): A dictionary to search for the entry.
        core_name_to_match (str): The core name to match in the dictionary values.

    Returns:
        dict: The dictionary entry matching the specified core name, or None if not found.
    """
    for key, value in format_dict.items():
        if value["core_name"] == core_name_to_match:
            return value



"""
includes a class that reads metadata from weather station
"""

from unidecode import unidecode


class MetaData:
    """
    A class for storing metadata information about a weather station.

    Attributes:
        station_name (str): The name of the weather station.
        station_id (str): The ID (DWD or DWD MOSMIX ID) of the weather station.
        altitude (float): The altitude of the weather station in meters.
        latitude (float): The latitude of the weather station in degree.
        longitude (float): The longitude of the weather station in degree.
        input_source (str): The source of input data for the station.
    """
    def __init__(self, **kwargs: str):
        self._station_name: str = "UnknownStationName"
        self.station_id: str = "UnknownStationID"
        self._altitude: float = None
        self._latitude: float = None
        self._longitude: float = None
        self.input_source: str = "UnknownInputSource"

        self.__dict__.update(kwargs)

    @property
    def station_name(self):
        return self._station_name

    @station_name.setter
    def station_name(self, value):
        """avoid special chars"""
        self._station_name = unidecode(value)

    @property
    def altitude(self) -> float:
        return self._altitude

    @altitude.setter
    def altitude(self, value: float) -> None:
        self._altitude = round(self._ensure_float(value), 5)

    @property
    def latitude(self) -> float:
        return self._latitude

    @latitude.setter
    def latitude(self, value: float) -> None:
        self._latitude = round(self._ensure_float(value), 5)

    @property
    def longitude(self) -> float:
        return self._longitude

    @longitude.setter
    def longitude(self, value: float) -> None:
        self._longitude = round(self._ensure_float(value), 5)

    def _ensure_float(self, value):
        if value is not None:
            try:
                return float(value)
            except:
                raise ValueError(f"Value must be of type float, not {type(value)}")



"""
converts core data to different simpler formats (currently without any transformation)
"""

import json
import pickle
import pandas as pd

from aixweather import definitions
from aixweather.imports.utils_import import MetaData


def to_pickle(
    core_df: pd.DataFrame,
    meta: MetaData,
    result_folder: str = None
):
    """Create and save a pickle file from the core data.

    Args:
        core_df (pd.DataFrame): DataFrame containing core data.
        meta (MetaData): Metadata associated with the data.

    Returns:
        pd.DataFrame: DataFrame containing the weather data formatted as core data.
    """

    core_df = core_df.copy()
    filename = f"Station_{meta.station_name}.pkl"
    meta_filename = f"Station_{meta.station_name}_meta_data.pkl"
    file_path = definitions.results_file_path(filename, result_folder)
    meta_file_path = definitions.results_file_path(meta_filename, result_folder)

    core_df.to_pickle(file_path)
    with open(meta_file_path, "wb") as file:
        pickle.dump(meta, file)

    print(f"Pickle saved to {file_path}, meta information saved to {meta_file_path}.")

    return core_df


def to_json(
    core_df: pd.DataFrame,
    meta: MetaData,
    result_folder: str = None
):
    """Create and save a json file from the core data.

    Args:
        core_df (pd.DataFrame): DataFrame containing core data.
        meta (MetaData): Metadata associated with the data.

    Returns:
        pd.DataFrame: DataFrame containing the weather data formatted as core data.
    """

    core_df = core_df.copy()

    filename = f"Station_{meta.station_name}.json"
    meta_filename = f"Station_{meta.station_name}_meta_data.json"
    file_path = definitions.results_file_path(filename, result_folder)
    meta_file_path = definitions.results_file_path(meta_filename, result_folder)

    # Convert DataFrame to JSON and save to file
    core_df.to_json(file_path, orient="records")

    # Convert meta_data to a dictionary and save to JSON file
    meta_dict = meta.__dict__  # Convert the meta_data object to a dictionary
    with open(meta_file_path, "w") as file:
        json.dump(meta_dict, file, indent=4)

    print(f"JSON saved to {file_path}, meta information saved to {meta_file_path}.")

    return core_df


def to_csv(
    core_df: pd.DataFrame,
    meta: MetaData,
    result_folder: str = None
):
    """Create and save a csv file from the core data.

    Args:
        core_df (pd.DataFrame): DataFrame containing core data.
        meta (MetaData): Metadata associated with the data.

    Returns:
        pd.DataFrame: DataFrame containing the weather data formatted as core data.
    """

    core_df = core_df.copy()

    filename = f"Station_{meta.station_name}.csv"
    meta_filename = f"Station_{meta.station_name}_meta_data.json"
    file_path = definitions.results_file_path(filename, result_folder)
    meta_file_path = definitions.results_file_path(meta_filename, result_folder)

    core_df.to_csv(file_path, index=True)

    # Convert meta_data to a dictionary and save to JSON file
    meta_dict = meta.__dict__  # Convert the meta_data object to a dictionary
    with open(meta_file_path, "w") as file:
        json.dump(meta_dict, file, indent=4)

    print(f"CSV saved to {file_path}, meta information saved to {meta_file_path}.")

    return core_df



"""
contains functions to handle data within ERC
"""

from tkinter import *
import json
import pandas as pd
from datetime import datetime

import requests

from aixweather.imports.utils_import import MetaData

key_list = [
    "4121.weatherstation.temperature",
    "4121.weatherstation.diffuse-radiation",
    "4121.weatherstation.global-radiation",
    "4121.weatherstation.pressure",
    "4121.weatherstation.rainfall",
    "4121.weatherstation.relative-humidity",
    "4121.weatherstation.wind-direction",
    "4121.weatherstation.wind-speed",
]


def load_credentials_ERC_weather_data() -> tuple:
    """
    GUI to return the username and password needed to access ERC weather data.

    Returns:
        tuple: The username and password required for accessing ERC weather data.
    """


    tk_window = Tk()
    tk_window.geometry("400x150")
    tk_window.title(
        "ERC is a private database, credentials are required to acquire the data"
    )

    # username label and text entry box
    username_label = Label(tk_window, text="User Name").grid(row=0, column=0)
    username = StringVar()
    username_entry = Entry(tk_window, textvariable=username).grid(row=0, column=1)

    # password label and password entry box
    password_label = Label(tk_window, text="Password").grid(row=1, column=0)
    password = StringVar()
    password_entry = Entry(tk_window, textvariable=password, show="*").grid(
        row=1, column=1
    )
    # login button
    login_button = Button(tk_window, text="Login", command=tk_window.destroy).grid(
        row=4, column=0
    )

    tk_window.mainloop()

    return (username.get(), password.get())


def import_ERC(start: datetime, stop: datetime, cred: tuple) -> pd.DataFrame:
    """
    Import weather data from aedifion ERC weather station (credentials required).

    Args:
        start (datetime): Datetime object representing the data start.
        stop (datetime): Datetime object representing the data stop.
        cred (tuple): Aedifion credentials.

    Returns:
        pd.DataFrame: DataFrame containing as raw as possible weather data from aedifion ERC.
    """

    # import cred if not given
    if cred is None:
        cred = load_credentials_ERC_weather_data()

    weather_df_tot = pd.DataFrame()

    for key in key_list:
        request = {
            "project_id": 11,
            "dataPointID": key,
            "start": start,
            "end": stop,
            "interpolation": "null",
            "samplerate": "1h",
            "short": "True",
        }
        response = requests.get(
            "https://api.ercebc.aedifion.io/v2/" + "datapoint/timeseries",
            auth=cred,
            params=request,
        )
        # Check for errors
        response.raise_for_status()

        # continue loading
        api_response = response.json()
        weather_df = _json2df(api_response, data_point_id=key)

        if not weather_df.empty:
            if weather_df_tot.empty:
                # If the main DataFrame is empty, simply assign the first DataFrame to it
                weather_df_tot = weather_df
            else:
                weather_df_tot = pd.merge(
                    weather_df_tot, weather_df, how="outer", left_index=True, right_index=True
                )

    return weather_df_tot


def import_meta_from_ERC() -> MetaData:
    """
    Import (specify) metadata from ERC.

    Returns:
        MetaData: A MetaData object containing station information for ERC.
    """

    meta = MetaData()
    meta.station_id = "ERC"
    meta.station_name = "Old_Experimental_Hall"
    meta.altitude = 230  # ca. 200m + building height 30m
    meta.latitude = 50.7893
    meta.longitude = 6.0516
    return meta


def _json2df(
    data: json = None, data_point_id: str = None, short: bool = True
) -> pd.DataFrame:
    """
    Transform the JSON returned by GET /v2/datapoint/timeseries to a Pandas DataFrame
    indexed by time.

    Args:
        data (json): The JSON data to be transformed into a DataFrame.
        data_point_id (str): Use dataPointID as the column name for values (default: value).
        short (bool): Whether the given JSON is in short format or not.

    Returns:
        pd.DataFrame: A Pandas DataFrame containing the transformed data.
    """
    if short:
        data = [[pd.to_datetime(t[0], utc=True), t[1]] for t in data]
    else:
        data = [[pd.to_datetime(t["time"], utc=True), t["value"]] for t in data["data"]]
    dataframe = pd.DataFrame(data, columns=["Time", data_point_id or "Value"])
    dataframe = dataframe.set_index("Time")
    return dataframe



"""
convert core data to epw (EnergyPlus) data
"""

import csv
import datetime as dt
import pandas as pd
import numpy as np

from aixweather import definitions
from aixweather.imports.utils_import import MetaData
from aixweather.transformation_functions import auxiliary, time_observation_transformations, pass_through_handling

"""
format_epw information:
for links see readme

Information for shifting:
Hour: This is the hour of the data. (1 - 24). Hour 1 is 00:01 to 01:00. Cannot be missing.
but e.g.:
DryBulbTemp: This is the dry bulb temperature in C at the time indicated.
and:
GlobHorRad: received on a horizontal surface during the hour preceding the time indicated.
----> Hence, we assume that hour 1 should show the DryBulbTemp from
0:30 to 1:30, i.e. the Temp at indicated time.

time of measurement checked by Martin Rätz (07.08.2023)
units checked by Martin Rätz (07.08.2023)
"""
format_epw = {
    "Year": {"core_name": "", "unit": "year", "time_of_meas_shift": None, "nan": None},
    "Month": {"core_name": "", "unit": "month", "time_of_meas_shift": None, "nan": None},
    "Day": {"core_name": "", "unit": "day", "time_of_meas_shift": None, "nan": None},
    "Hour": {"core_name": "", "unit": "hour", "time_of_meas_shift": None, "nan": None},
    "Minute": {"core_name": "", "unit": "minute", "time_of_meas_shift": None, "nan": None},
    "Data Source and Uncertainty Flags": {"core_name": "", "unit": None, "time_of_meas_shift": None, "nan": "?"},
    "DryBulbTemp": {"core_name": "DryBulbTemp", "unit": "degC", "time_of_meas_shift": None, "nan": 99.9},
    "DewPointTemp": {"core_name": "DewPointTemp", "unit": "degC", "time_of_meas_shift": None, "nan": 99.9},
    "RelHum": {"core_name": "RelHum", "unit": "percent", "time_of_meas_shift": None, "nan": 999.0},
    "AtmPressure": {"core_name": "AtmPressure", "unit": "Pa", "time_of_meas_shift": None, "nan": 999999.0},
    "ExtHorRad": {"core_name": "ExtHorRad", "unit": "Wh/m2", "time_of_meas_shift": 'ind2prec', "nan": 9999.0},
    "ExtDirNormRad": {"core_name": "ExtDirNormRad", "unit": "Wh/m2", "time_of_meas_shift": 'ind2prec', "nan": 9999.0},
    "HorInfra": {"core_name": "HorInfra", "unit": "Wh/m2", "time_of_meas_shift": 'ind2prec', "nan": 9999.0},
    "GlobHorRad": {"core_name": "GlobHorRad", "unit": "Wh/m2", "time_of_meas_shift": 'ind2prec', "nan": 9999.0},
    "DirNormRad": {"core_name": "DirNormRad", "unit": "Wh/m2", "time_of_meas_shift": 'ind2prec', "nan": 9999.0},
    "DiffHorRad": {"core_name": "DiffHorRad", "unit": "Wh/m2", "time_of_meas_shift": 'ind2prec', "nan": 9999.0},
    "GlobHorIll": {"core_name": "GlobHorIll", "unit": "lux", "time_of_meas_shift": 'ind2prec', "nan": 999999.0},
    "DirecNormIll": {"core_name": "DirecNormIll", "unit": "lux", "time_of_meas_shift": 'ind2prec', "nan": 999999.0},
    "DiffuseHorIll": {"core_name": "DiffuseHorIll", "unit": "lux", "time_of_meas_shift": 'ind2prec', "nan": 999999.0},
    "ZenithLum": {"core_name": "ZenithLum", "unit": "Cd/m2", "time_of_meas_shift": 'ind2prec', "nan": 9999.0},
    "WindDir": {"core_name": "WindDir", "unit": "deg", "time_of_meas_shift": None, "nan": 999.0},
    "WindSpeed": {"core_name": "WindSpeed", "unit": "m/s", "time_of_meas_shift": None, "nan": 999.0},
    "TotalSkyCover": {"core_name": "TotalSkyCover", "unit": "1tenth", "time_of_meas_shift": None, "nan": 99},
    "OpaqueSkyCover": {"core_name": "OpaqueSkyCover", "unit": "1tenth", "time_of_meas_shift": None, "nan": 99},
    "Visibility": {"core_name": "Visibility", "unit": "km", "time_of_meas_shift": None, "nan": 9999.0},
    "CeilingH": {"core_name": "CeilingH", "unit": "m", "time_of_meas_shift": None, "nan": 99999},
    "WeatherObs": {"core_name": "", "unit": "None", "time_of_meas_shift": None, "nan": 9},
    "WeatherCode": {"core_name": "", "unit": "None", "time_of_meas_shift": None, "nan": 999999999},
    "PrecWater": {"core_name": "PrecWater", "unit": "mm", "time_of_meas_shift": None, "nan": 999.0},
    "Aerosol": {"core_name": "Aerosol", "unit": "1thousandth", "time_of_meas_shift": None, "nan": 0.999},
    "Snow": {"core_name": "", "unit": "cm", "time_of_meas_shift": None, "nan": 999.0},
    "DaysSinceSnow": {"core_name": "", "unit": "days", "time_of_meas_shift": None, "nan": 99},
    "Albedo": {"core_name": "", "unit": "None", "time_of_meas_shift": None, "nan": 999},
    "LiquidPrecD": {"core_name": "LiquidPrecD", "unit": "mm/h", "time_of_meas_shift": None, "nan": 999},
    "LiquidPrepQuant": {"core_name": "", "unit": "hours", "time_of_meas_shift": None, "nan": 99},
}


def to_epw(
    core_df: pd.DataFrame,
    meta: MetaData,
    start: dt.datetime,
    stop: dt.datetime,
    fillna: bool,
    result_folder: str = None
) -> pd.DataFrame:
    """Create an EPW file from the core data.

    Args:
        core_df (pd.DataFrame): DataFrame containing core data.
        meta (MetaData): Metadata associated with the weather data.
        start (dt.datetime): Timestamp for the start of the EPW file.
        stop (dt.datetime): Timestamp for the end of the EPW file.
        fillna (bool): Boolean indicating whether NaN values should be filled.

    Returns:
        pd.DataFrame: DataFrame containing the weather data formatted for EPW export,
                      excluding metadata.
    """

    ### evaluate correctness of format
    auxiliary.evaluate_transformations(
        core_format=definitions.format_core_data, other_format=format_epw
    )

    df = core_df.copy()

    filename = (
        f"{meta.station_id}_{start.strftime('%Y%m%d')}_{stop.strftime('%Y%m%d')}"
        f"_{meta.station_name}.epw"
    )
    # get file path to safe data to
    file_path = definitions.results_file_path(filename, result_folder)

    ### create header lines
    def line1_location(
        meta: MetaData,
    ):
        """
        Get location metadata (station name, state, country, data_type,
        stationID, lat, lon, TZ, alt)

        return:
            location:      List        Erstezeile(LOCATION) von epw Daten als List
        """

        data_type = ""
        timezone = 0  # relative to UTC

        location = [
            "LOCATION",
            meta.station_name,
            "State",
            "country",
            data_type,
            meta.station_id,
            str(meta.latitude),
            str(meta.longitude),
            timezone,
            str(meta.altitude),
        ]

        return location

    def line2_design_cond():
        """
        Erstellen zweite Zeile der epw.

        return:
            design_cond:    List    Zweite Zeile(Design Condition) von epw Daten als List
        """
        design_cond = [
            "DESIGN CONDITIONS",
            0,  # number of design condition
        ]

        return design_cond

    def line3_typ_ext_period(df):
        """
        Parsen von weatherdata um typische und extreme Perioden zu holen.

        Typische Perioden sind Wochen mit Temperatur, die der Durchschnittstemperatur der Saison
        am nächsten kommt.
        Extreme Perioden sind Wochen mit Temperatur, die der Maximum-/Minimumtemperatur der Saison
        am nächsten kommt.

        return:
            typical_extreme_period:     List    Dritte Zeile(TYPICAL/EXTREME PERIODS)
                                                    von epw Daten als List
        """

        typical_extreme_period = [
            "TYPICAL/EXTREME PERIODS",
        ]

        season_dict = {
            11: "Autumn",
            12: "Winter",
            1: "Winter",
            2: "Winter",
            3: "Spring",
            4: "Spring",
            5: "Spring",
            6: "Summer",
            7: "Summer",
            8: "Summer",
            9: "Autumn",
            10: "Autumn",
        }  # Monaten in Saisons zuweisen

        def group_func(input):
            """Gruppefunktion für df.groupby()"""
            return season_dict[input.month]

        df_temp_ambient = df["DryBulbTemp"]  # Temperature_Ambient von weatherdata holen
        number_of_periods = (
            df_temp_ambient.groupby(group_func).mean().shape[0]
        )  # Zahl von der Saisons rechnen als Zahl von Perioden
        typical_extreme_period.append(number_of_periods)

        # Gruppierung per Saison
        try:
            summer_temp = df_temp_ambient.groupby(group_func).get_group("Summer")
        except KeyError:
            summer_temp = pd.DataFrame()
        try:
            spring_temp = df_temp_ambient.groupby(group_func).get_group("Spring")
        except KeyError:
            spring_temp = pd.DataFrame()
        try:
            autumn_temp = df_temp_ambient.groupby(group_func).get_group("Autumn")
        except KeyError:
            autumn_temp = pd.DataFrame()
        try:
            winter_temp = df_temp_ambient.groupby(group_func).get_group("Winter")
        except KeyError:
            winter_temp = pd.DataFrame()

        if not summer_temp.empty:
            typical_extreme_period[1] = (
                typical_extreme_period[1] + 1
            )  # Summer und Winter haben extreme Periode.
            max_temp_summer = summer_temp.max()
            typ_temp_summer = summer_temp.mean()
            summer_temp_w = summer_temp.resample(
                "W", label="left"
            ).mean()  # Resample in wochentliche Interval

            # Datenpunkt(typisch und extreme) finden
            idx_mean_summer = summer_temp_w.sub(typ_temp_summer).abs().idxmin()
            idx_max_summer = summer_temp_w.sub(max_temp_summer).abs().idxmin()
            week_closest2mean_summer = summer_temp_w.loc[[idx_mean_summer]]  # Starttag
            week_closest2max_summer = summer_temp_w.loc[[idx_max_summer]]  # Starttag

            # Endtag berechnen
            weekend_max_summer = week_closest2max_summer.index + dt.timedelta(days=6)
            weekend_mean_summer = week_closest2mean_summer.index + dt.timedelta(days=6)

            # List für die Saison erstellen
            summer = [
                "Summer - Week Nearest Max Temperature For Period",
                "Extreme",
                str(week_closest2max_summer.index.month[0])
                + "/"
                + str(week_closest2max_summer.index.day[0]),
                str(weekend_max_summer.month[0]) + "/" + str(weekend_max_summer.day[0]),
                "Summer - Week Nearest Average Temperature For Period",
                "Typical",
                str(week_closest2mean_summer.index.month[0])
                + "/"
                + str(week_closest2mean_summer.index.day[0]),
                str(weekend_mean_summer.month[0])
                + "/"
                + str(weekend_mean_summer.day[0]),
            ]

            typical_extreme_period = (
                typical_extreme_period + summer
            )  # Liste zusammensetzen

        # für alle Saison wiederholen
        if not winter_temp.empty:
            typical_extreme_period[1] = typical_extreme_period[1] + 1
            min_temp_winter = winter_temp.min()
            typ_temp_winter = winter_temp.mean()
            winter_temp_w = winter_temp.resample("W", label="left").mean()
            idx_mean_winter = winter_temp_w.sub(typ_temp_winter).abs().idxmin()
            idx_min_winter = winter_temp_w.sub(min_temp_winter).abs().idxmin()
            week_closest2mean_winter = winter_temp_w.loc[[idx_mean_winter]]
            week_closest2min_winter = winter_temp_w.loc[[idx_min_winter]]
            weekend_min_winter = week_closest2min_winter.index + dt.timedelta(days=6)
            weekend_mean_winter = week_closest2mean_winter.index + dt.timedelta(days=6)
            winter = [
                "Winter - Week Nearest Min Temperature For Period",
                "Extreme",
                str(week_closest2min_winter.index.month[0])
                + "/"
                + str(week_closest2min_winter.index.day[0]),
                str(weekend_min_winter.month[0]) + "/" + str(weekend_min_winter.day[0]),
                "Winter - Week Nearest Average Temperature For Period",
                "Typical",
                str(week_closest2mean_winter.index.month[0])
                + "/"
                + str(week_closest2mean_winter.index.day[0]),
                str(weekend_mean_winter.month[0])
                + "/"
                + str(weekend_mean_winter.day[0]),
            ]

            typical_extreme_period = typical_extreme_period + winter

        if not autumn_temp.empty:
            typ_temp_autumn = autumn_temp.mean()
            autumn_temp_w = autumn_temp.resample("W", label="left").mean()
            idx_mean_autumn = autumn_temp_w.sub(typ_temp_autumn).abs().idxmin()
            week_closest2mean_autumn = autumn_temp_w.loc[[idx_mean_autumn]]
            weekend_mean_autumn = week_closest2mean_autumn.index + dt.timedelta(days=6)
            autumn = [
                "Autumn - Week Nearest Average Temperature For Period",
                "Typical",
                str(week_closest2mean_autumn.index.month[0])
                + "/"
                + str(week_closest2mean_autumn.index.day[0]),
                str(weekend_mean_autumn.month[0])
                + "/"
                + str(weekend_mean_autumn.day[0]),
            ]

            typical_extreme_period = typical_extreme_period + autumn

        if not spring_temp.empty:
            typ_temp_spring = spring_temp.mean()
            spring_temp_w = spring_temp.resample("W", label="left").mean()
            idx_mean_spring = spring_temp_w.sub(typ_temp_spring).abs().idxmin()
            week_closest2mean_spring = spring_temp_w.loc[[idx_mean_spring]]
            weekend_mean_spring = week_closest2mean_spring.index + dt.timedelta(days=6)
            spring = [
                "Spring - Week Nearest Average Temperature For Period",
                "Typical",
                str(week_closest2mean_spring.index.month[0])
                + "/"
                + str(week_closest2mean_spring.index.day[0]),
                str(weekend_mean_spring.month[0])
                + "/"
                + str(weekend_mean_spring.day[0]),
            ]

            typical_extreme_period = typical_extreme_period + spring

        return typical_extreme_period

    def line4_ground_temp(df):
        """
        Parsen von weatherdata, um Bodentemperaturen zu holen

        return:
            ground_temp:    List    Vierte Zeile(GROUND TEMPERATURES) von epw Daten als List
        """

        ground_temp = [
            "GROUND TEMPERATURES",
        ]

        df_w_ground = (
            df.resample("M").mean().copy().round(decimals=1)
        )  # Resample in monatliche Interval
        try:
            ground_t = df_w_ground[
                [
                    "Soil_Temperature_5cm",
                    "Soil_Temperature_10cm",
                    "Soil_Temperature_20cm",
                    "Soil_Temperature_50cm",
                    "Soil_Temperature_1m",
                ]
            ].to_numpy()  # Dataframe2Array
            # Array zu Liste umwandeln -> Zusammensetzen
            ground_temp = (
                ground_temp
                + [5]  # ground layers
                + [0.05, None, None, None]
                + ground_t[:, 0].tolist()
                + [0.1, None, None, None]
                + ground_t[:, 1].tolist()
                + [0.2, None, None, None]
                + ground_t[:, 2].tolist()
                + [0.5, None, None, None]
                + ground_t[:, 3].tolist()
                + [1, None, None, None]
                + ground_t[:, 4].tolist()
            )
            return ground_temp
        except KeyError as err:
            print(
                f"For adding the probably unnecessary ground temperature to the .epw file header, "
                f"the following made it impossible: {err}"
            )
            ground_temp = ground_temp + [0]  # 0 ground layers

            return ground_temp

    def line5_holiday_dl_saving(df):
        """
        Erstellen der 5. Zeile der epw.

        return:
            holiday_dl_saving:    List    5.Zeile(HOLIDAYS/DAYLIGHT SAVINGS) von epw Daten als List
        """

        if True in df.index.is_leap_year:
            isLeap = "Yes"
        else:
            isLeap = "No"
        holiday_dl_saving = [
            "HOLIDAYS/DAYLIGHT SAVINGS",
            isLeap,  # Leap Year Observed
            0,  # Daylight Saving Start Date
            0,  # Daylight Saving End Date
            0,  # Number of Holidays
        ]
        return holiday_dl_saving

    def line6_comment_1():
        """
        Erstellen der 6. Zeile der epw.

        return:
            comment_1:    List    6.Zeile(COMMENTS 1) von epw Daten als List
        """
        return [
            "COMMENTS 1",
            "For data format information see the code or check: "
            "https://designbuilder.co.uk/cahelp/Content/EnergyPlusWeatherFileFormat.htm",
        ]

    def line7_comment_2(comment2=None):
        """
        Erstellen der 7. Zeile der epw.

        return:
            comment_2:    List    7.Zeile(COMMENTS 2) von epw Daten als List
        """
        return ["COMMENTS 2", comment2]

    def line8_data_periods(df):
        """
        Parsen von weatherdata, um Start- und Enddatenpunkt zu holen

        return:
            data_periods:    List    8.Zeile(DATA PERIODS) von epw Daten als List
        """
        start_dp = df.index[0]
        end_dp = df.index[-2]
        data_periods = [
            "DATA PERIODS",
            1,  # Anzahl von Datenperioden
            1,  # Anzahl von Intervale in einer Stunde
            "Data",  # DP Name oder Beschreibung
            start_dp.strftime("%A"),  # DP Starttag
            start_dp.strftime("%m/%d"),  # DP Startdatum
            end_dp.strftime("%m/%d"),  # DP Enddatum
        ]
        return data_periods

    ### parse actual data
    def format_data(df, start, stop):
        """
        Parsen von weatherdata, für den export

        return:
            data_list:    List    Datasätze von epw Daten als List
        """

        # add 1 hour to start, because epw always starts at hour 1 instead of 0
        start_epw = start + dt.timedelta(hours=1)

        ### measurement time conversion
        df = time_observation_transformations.shift_time_by_dict(format_epw, df)

        ### if possible avoid back and forth interpolating -> pass through
        ### variables without shifting
        df = pass_through_handling.pass_through_measurements_with_back_and_forth_interpolating(
           df, format_epw
        )

        ### select only desired period
        df = time_observation_transformations.truncate_data_from_start_to_stop(
            df, start_epw, stop
        )

        ### select the desired columns
        df = auxiliary.force_data_variable_convention(df, format_epw)

        # fill newly created variables of desired output format
        # Index von Dataframe aufspalten
        df["Year"] = pd.DatetimeIndex(df.index).year
        df["Month"] = pd.DatetimeIndex(df.index).month
        df["Day"] = pd.DatetimeIndex(df.index).day
        df["Hour"] = pd.DatetimeIndex(df.index).hour
        df["Minute"] = pd.DatetimeIndex(df.index).minute

        ### meet special epw requirements
        # Stunden 0 zu 24 der vorherigen Tag umwandeln
        df["Hour"] = df["Hour"].replace([0], 24)
        # Falls Tag ungleich 1 -> Tag substrahieren mit 1
        df.loc[(df["Hour"] == 24) & (df["Day"] != 1), "Day"] = df.loc[
            (df["Hour"] == 24) & (df["Day"] != 1), "Day"
        ].sub(1)
        # Falls Tag gleich 1 -> Jahr, Monat, Tag loeschen -> mit ffill nachfuellen
        df.loc[
            (df["Hour"] == 24) & (df["Day"] == 1),
            ["Year", "Month", "Day"]
        ] = np.nan
        df["Year"] = (
            df["Year"].ffill().bfill().astype(int)
        )
        df["Month"] = (
            df["Month"].ffill().bfill().astype(int)
        )
        df["Day"] = df["Day"].ffill().bfill().astype(int)
        df.reset_index(drop=True, inplace=True)

        # data should always contain full days
        def fill_full_first_day(df):
            # Identify the first hour and date of the DataFrame
            first_minute = df.iloc[0]["Minute"]
            first_hour = df.iloc[0]["Hour"]
            first_day = df.iloc[0]["Day"]
            first_month = df.iloc[0]["Month"]
            first_year = df.iloc[0]["Year"]
            rows_to_add = 0

            # If the first hour is not 1, add rows to start with hour 1
            if first_hour != 1:
                # Calculate how many rows to add
                rows_to_add = int(first_hour) - 1

                # Generate new rows
                for i in range(rows_to_add, 0, -1):
                    new_row = pd.DataFrame(
                        {
                            "Minute": [first_minute],
                            "Hour": [i],
                            "Day": [first_day],
                            "Month": [first_month],
                            "Year": [first_year],
                        }
                    )
                    df = pd.concat([new_row, df]).reset_index(drop=True)
            return df, rows_to_add

        def fill_full_last_day(df):
            # Identify the last hour and date of the DataFrame
            last_hour = df.iloc[-1]["Hour"]
            last_day = df.iloc[-1]["Day"]
            last_month = df.iloc[-1]["Month"]
            last_year = df.iloc[-1]["Year"]
            last_minute = df.iloc[-1]["Minute"]
            rows_to_add = 0

            # If the last hour is not 24, add rows to reach hour 24
            if last_hour != 24:
                # Calculate how many rows to add
                rows_to_add = 24 - int(last_hour)

                # Generate new rows
                new_rows = []
                for i in range(1, rows_to_add + 1):
                    new_row = {
                        "Minute": last_minute,
                        "Hour": last_hour + i,
                        "Day": last_day,
                        "Month": last_month,
                        "Year": last_year,
                    }
                    new_rows.append(new_row)

                # Append new rows to DataFrame
                df = pd.concat([df, pd.DataFrame(new_rows)], ignore_index=True)
            return df, rows_to_add

        df, first_day_added_rows = fill_full_first_day(df)
        df, last_day_added_rows = fill_full_last_day(df)

        # ensure data type where required
        columns_to_convert = ["Year", "Month", "Day", "Hour", "Minute"]
        for col in columns_to_convert:
            df[col] = df[col].astype(float).astype(int)

        ### fill NaNs
        if fillna:
            # Forward-fill added rows at end of df
            df.iloc[-last_day_added_rows:, :] = df.ffill().iloc[
                -last_day_added_rows:, :
                                                      ]
            # fill added rows at beginning of df
            df.iloc[:first_day_added_rows, :] = df.bfill().iloc[
                :first_day_added_rows, :
                                                      ]

            # fill first and last lines nans (possibly lost through shifting)
            df.iloc[0 + first_day_added_rows + 1, :] = df.bfill().iloc[
                0 + first_day_added_rows + 1, :
                                                             ]
            df.iloc[-1 - last_day_added_rows, :] = df.ffill().iloc[
                -1 - last_day_added_rows, :
                                                         ]

            # fill default nans to the rest
            df = auxiliary.fill_nan_from_format_dict(df, format_epw)

        # cut off float digits (required for EnergyPlus)
        df = df.map(lambda x: (f"{x:.1f}") if isinstance(x, float) else x)

        # again make sure correct order and variables are applied
        # (processing might have mixed it up)
        df = auxiliary.force_data_variable_convention(df, format_epw)

        ### format dataframe to list
        data_list = df[format_epw.keys()].to_numpy().tolist()

        return data_list, df

    ### merge all header lines and the data to be saved in a .epw file
    with open(file_path, "w", newline="", encoding="latin1") as file:
        writer = csv.writer(file)
        writer.writerows(
            [
                line1_location(meta),
                line2_design_cond(),
                line3_typ_ext_period(df),
                line4_ground_temp(df),
                line5_holiday_dl_saving(df),
                line6_comment_1(),
                line7_comment_2(),
                line8_data_periods(df),
            ]
        )
        df_as_list, df = format_data(df, start, stop)
        writer.writerows(df_as_list)

    print(f"EPW file saved to {file_path}.")

    return df



"""
This module includes variables or functions which define central variables for the whole project.
"""

import os

# format of the core_data
# time of measurement is always at indicated time
format_core_data = {
    # from TMY3 https://www.nrel.gov/docs/fy08osti/43156.pdf
    "DryBulbTemp": {"unit": "degC"},
    "DewPointTemp": {"unit": "degC"},
    "RelHum": {"unit": "percent"},
    "ExtHorRad": {"unit": "Wh/m2"},
    "ExtDirNormRad": {"unit": "Wh/m2"},
    "HorInfra": {"unit": "Wh/m2"},
    "GlobHorRad": {"unit": "Wh/m2"},
    "DirNormRad": {"unit": "Wh/m2"},
    "DirHorRad": {"unit": "Wh/m2"},
    "DiffHorRad": {"unit": "Wh/m2"},
    "GlobHorIll": {"unit": "lux"},
    "DirecNormIll": {"unit": "lux"},
    "DiffuseHorIll": {"unit": "lux"},
    "ZenithLum": {"unit": "Cd/m2"},
    "WindDir": {"unit": "deg"},
    "WindSpeed": {"unit": "m/s"},
    "TotalSkyCover": {"unit": "1tenth"},
    "OpaqueSkyCover": {"unit": "1tenth"},
    "Visibility": {"unit": "km"},
    "CeilingH": {"unit": "m"},
    "PrecWater": {"unit": "mm"},
    "Aerosol": {"unit": "1thousandth"},
    "LiquidPrecD": {"unit": "mm/h"},
    # exception to TMY3 format as all TMY3 data file actually use "Pa" instead of mbar
    "AtmPressure": {"unit": "Pa"},
    # additional variables
    "Soil_Temperature_5cm": {"unit": "degC"},
    "Soil_Temperature_10cm": {"unit": "degC"},
    "Soil_Temperature_20cm": {"unit": "degC"},
    "Soil_Temperature_50cm": {"unit": "degC"},
    "Soil_Temperature_1m": {"unit": "degC"},
}

# creates the path to the root directory to be used in other modules
ROOT_DIR = os.path.realpath(os.path.join(os.path.dirname(__file__), ".."))

# local temporary folder
local_folder_temp = os.path.join(ROOT_DIR, "temp_local")

def result_folder_path() -> str:
    """
    Creates the path to the resultsfolder.
    """
    folder_path = os.path.join(f"{ROOT_DIR}", "results")
    return folder_path

def results_file_path(filename: str, folder_path: str = None) -> str:
    """
    create a path to a results file

    Args:
        filename: name of file
        folder_path: path to result folder

    Returns:
        str: path to result file
    """
    # if required specify folder_path e.g. for Django WebApp
    if folder_path is None:
        folder_path = result_folder_path()

    # Ensure that the results folder exists
    os.makedirs(folder_path, exist_ok=True)

    # create file path
    filepath = os.path.join(folder_path, filename)
    return filepath



"""
This module includes functions to transform ERC data to core data format.
"""

import pandas as pd

from aixweather import definitions
from aixweather.imports.utils_import import MetaData
from aixweather.transformation_functions import auxiliary, time_observation_transformations, variable_transformations, \
    pass_through_handling, unit_conversions

"""
format_ERC information

checked by Martin Rätz 01.09.2023
"""
format_ERC = {
    '4121.weatherstation.temperature': {'core_name': 'DryBulbTemp', 'time_of_meas_shift': 'foll2ind', 'unit': "degC"},
    '4121.weatherstation.diffuse-radiation': {'core_name': 'DiffHorRad', 'time_of_meas_shift': 'foll2ind', 'unit': "Wh/m2"},
    '4121.weatherstation.global-radiation': {'core_name': 'GlobHorRad', 'time_of_meas_shift': 'foll2ind', 'unit': "Wh/m2"},
    '4121.weatherstation.pressure': {'core_name': 'AtmPressure', 'time_of_meas_shift': 'foll2ind', 'unit': "hPa"},
    '4121.weatherstation.relative-humidity': {'core_name': 'RelHum', 'time_of_meas_shift': 'foll2ind', 'unit': "percent"},
    '4121.weatherstation.wind-direction': {'core_name': 'WindDir', 'time_of_meas_shift': 'foll2ind', 'unit': "deg"},
    '4121.weatherstation.wind-speed': {'core_name': 'WindSpeed', 'time_of_meas_shift': 'foll2ind', 'unit': "m/s"},
    # too little information available, also possibly needs a sum up not a mean when interpolating
    # 'rainfall': {'core_name': 'PrecWater',
    # 'time_of_meas_shift': 'foll2ind',
    # 'unit': "mm/h"}
}


def ERC_to_core_data(df_import: pd.DataFrame, meta: MetaData) -> pd.DataFrame:
    """
    Transform imported ERC (Energy Research Center) weather data into core data format.

    Args:
        df_import (pd.DataFrame): The DataFrame containing imported ERC weather data.
        meta (MetaData): Metadata associated with the data.

    Returns:
        pd.DataFrame: The transformed DataFrame in the core data format.
    """

    ### evaluate correctness of format
    auxiliary.evaluate_transformations(
        core_format=definitions.format_core_data, other_format=format_ERC
    )

    ### format raw data for further operations
    df = df_import.copy()
    # Resample the DataFrame to make the DatetimeIndex complete and monotonic
    df = df.resample('H').asfreq()
    # Remove timezone awareness
    df.index = df.index.tz_localize(None)
    # rename available variables to core data format
    df = auxiliary.rename_columns(df, format_ERC)

    ### convert timezone to UTC
    # data is UTC

    ### shift and interpolate data forward 30mins or backward -30mins
    df_no_shift = df.copy()
    df = time_observation_transformations.shift_time_by_dict(format_ERC, df)

    def transform_ERC(df):
        # drop unnecessary variables
        df = auxiliary.force_data_variable_convention(df, definitions.format_core_data)

        ### convert units
        df["AtmPressure"] = unit_conversions.hPa_to_Pa(df["AtmPressure"])

        ### impute missing variables from other available ones
        df, calc_overview = variable_transformations.variable_transform_all(df, meta)

        return df, calc_overview

    df, meta.executed_transformations = transform_ERC(df)

    ### add unshifted data for possible later direct use (pass-through),
    ### to avoid back and forth interpolating
    df = pass_through_handling.create_pass_through_variables(
        df_shifted=df,
        df_no_shift=df_no_shift,
        format=format_ERC,
        transform_func=transform_ERC,
        meta=meta,
    )

    return df



"""
import epw files
"""

import pandas as pd

from aixweather.imports.utils_import import MetaData


def load_epw_meta_from_file(path: str) -> MetaData:
    """
    Load an EPW file from a specified path and parse the file header for metadata.

    Args:
        path (str): The file path to the EPW file to be loaded.

    Returns:
        MetaData: An object containing the parsed metadata from the EPW file.
    """

    meta = MetaData()

    with open(path, "r", encoding="latin1") as file:
        lines = file.readlines()

    # The 1st line contains the location data
    location_data = lines[0].split(",")

    meta.station_name = location_data[1]
    meta.latitude = float(location_data[6])
    meta.longitude = float(location_data[7])
    meta.timezone = float(location_data[8])
    meta.altitude = float(location_data[9])
    meta.input_source = "EPW"

    return meta


def load_epw_from_file(path: str) -> pd.DataFrame:
    """
    Import data from an EPW file and convert it into a DataFrame.

    Args:
        path (str): The absolute path to the EPW file.

    Returns:
        pd.DataFrame: A DataFrame containing the imported data from the EPW file.
    """


    # Find the row number for "DATA PERIODS" to determine where the data starts
    with open(path, "r", encoding="latin1") as file:
        lines = file.readlines()
        data_start_row = (
                next(i for i, line in enumerate(lines) if "DATA PERIODS" in line) + 1
        )

    # Skipping header rows to load data
    weather_df = pd.read_csv(
        path,
        skiprows=data_start_row,
        header=None,
        encoding="latin1",
        encoding_errors="replace",
    )

    # The first 4 columns represent year, month, day, and hour respectively,
    # but with hour 24 instead of hour 0.
    hour = weather_df.iloc[:, 3].copy()
    mask_24hr = hour == 24
    hour.loc[mask_24hr] = 0
    weather_df["datetime"] = pd.to_datetime(
        weather_df.iloc[:, 0].astype(str)
        + "-"
        + weather_df.iloc[:, 1].astype(str)
        + "-"
        + weather_df.iloc[:, 2].astype(str)
        + " "
        + hour.astype(str)
        + ":00:00"
    )

    # Increment the day by one for those rows where hour
    # was originally 24
    weather_df.loc[mask_24hr, "datetime"] = weather_df.loc[mask_24hr, "datetime"] \
                                            + pd.Timedelta(days=1)

    # Setting datetime column as index
    weather_df.set_index("datetime", inplace=True)

    return weather_df



import pandas as pd

from aixweather import definitions
from aixweather.imports.utils_import import MetaData
from aixweather.transformation_functions import auxiliary, time_observation_transformations, variable_transformations, \
    pass_through_handling, unit_conversions

"""
format_TRY_15_45 information 

checked by Martin Rätz (08.08.2023)
"""
format_TRY_15_45 = {
    "t": {"core_name": "DryBulbTemp", "time_of_meas_shift": "prec2ind", "unit": "degC"},
    "p": {"core_name": "AtmPressure", "time_of_meas_shift": None, "unit": "hPa"},
    "WR": {"core_name": "WindDir", "time_of_meas_shift": None, "unit": "deg"},
    "WG": {"core_name": "WindSpeed", "time_of_meas_shift": None, "unit": "m/s"},
    "N": {"core_name": "TotalSkyCover", "time_of_meas_shift": None, "unit": "1eigth"},
    # 'x',
    "RF": {"core_name": "RelHum", "time_of_meas_shift": "prec2ind", "unit": "percent"},
    "B": {"core_name": "DirHorRad", "time_of_meas_shift": "prec2ind", "unit": "Wh/m2"},
    "D": {"core_name": "DiffHorRad", "time_of_meas_shift": "prec2ind", "unit": "Wh/m2"},
    "A": {"core_name": "HorInfra", "time_of_meas_shift": None, "unit": "Wh/m2"},
    # 'E',
}


def TRY_to_core_data(df_import: pd.DataFrame, meta: MetaData) -> pd.DataFrame:
    """
    Transform imported TRY data of the formats 2015 and 2045 into the core data format.

    Args:
        df_import (pd.DataFrame): The DataFrame containing imported TRY weather data.
        meta (MetaData): Metadata associated with the data.

    Returns:
        pd.DataFrame: The transformed DataFrame in the core data format.
    """

    ### evaluate correctness of format
    auxiliary.evaluate_transformations(
        core_format=definitions.format_core_data, other_format=format_TRY_15_45
    )

    ### preprocessing raw data for further operations
    df = df_import.copy()
    # Resample the DataFrame to make the DatetimeIndex complete and monotonic
    df = df.resample('H').asfreq()
    # rename available variables to core data format
    df = auxiliary.rename_columns(df, format_TRY_15_45)

    ### convert timezone to UTC+0
    df = df.shift(periods=-1, freq="H", axis=0)

    ### shift and interpolate data forward 30mins or backward -30mins
    df_no_shift = df.copy()
    df = time_observation_transformations.shift_time_by_dict(format_TRY_15_45, df)

    def transform_TRY(df):
        # drop unnecessary variables
        df = auxiliary.force_data_variable_convention(df, definitions.format_core_data)
        ### convert units
        df["TotalSkyCover"] = unit_conversions.eigth_to_tenth(df["TotalSkyCover"])
        df["AtmPressure"] = unit_conversions.hPa_to_Pa(df["AtmPressure"])
        ### impute missing variables from other available ones
        df, calc_overview = variable_transformations.variable_transform_all(df, meta)
        return df, calc_overview

    df, meta.executed_transformations = transform_TRY(df)

    ### add unshifted data for possible later direct use (pass-through),
    ### to avoid back and forth interpolating
    df = pass_through_handling.create_pass_through_variables(
        df_shifted=df,
        df_no_shift=df_no_shift,
        format=format_TRY_15_45,
        transform_func=transform_TRY,
        meta=meta,
    )

    return df



from datetime import datetime, timedelta
import pandas as pd

from aixweather import definitions
from aixweather.imports.utils_import import MetaData
from aixweather.transformation_functions import auxiliary, time_observation_transformations, variable_transformations, \
    pass_through_handling, unit_conversions

"""
format_DWD_historical information:
see readme

checked by Martin Rätz (08.08.2023)
"""
format_DWD_historical = {
    # https://opendata.dwd.de/climate_environment/CDC/observations_germany/climate/
    # 10_minutes/air_temperature/DESCRIPTION_obsgermany_climate_10min_air_temperature_en.pdf
    "RF_10": {"core_name": "RelHum", "time_of_meas_shift": "foll2ind", "unit": "percent"},
    "TT_10": {"core_name": "DryBulbTemp", "time_of_meas_shift": "foll2ind", "unit": "degC"},
    "TD_10": {"core_name": "DewPointTemp", "time_of_meas_shift": "foll2ind", "unit": "degC"},
    "PP_10": {"core_name": "AtmPressure", "time_of_meas_shift": "foll2ind", "unit": "hPa"},
    #https://opendata.dwd.de/climate_environment/CDC/observations_germany/climate/10_minutes/solar/BESCHREIBUNG_obsgermany_climate_10min_solar_de.pdf
    "DS_10": {"core_name": "DiffHorRad", "time_of_meas_shift": "foll2ind", "unit": "J/cm2", "resample": "sum"},
    #https://de.wikipedia.org/wiki/Globalstrahlung
    "GS_10": {"core_name": "GlobHorRad", "time_of_meas_shift": "foll2ind", "unit": "J/cm2", "resample": "sum"},
    #https://de.wikipedia.org/wiki/Atmosph%C3%A4rische_Gegenstrahlung
    "LS_10": {"core_name": "HorInfra", "time_of_meas_shift": "foll2ind", "unit": "J/cm2", "resample": "sum", "nan":[990, -999]},
    #https://opendata.dwd.de/climate_environment/CDC/observations_germany/climate/10_minutes/wind/BESCHREIBUNG_obsgermany_climate_10min_wind_de.pdf
    "FF_10": {"core_name": "WindSpeed", "time_of_meas_shift": "prec2ind", "unit": "m/s"},
    "DD_10": {"core_name": "WindDir", "time_of_meas_shift": "prec2ind", "unit": "deg"},
    # https://opendata.dwd.de/climate_environment/CDC/observations_germany/climate/10_minutes/precipitation/BESCHREIBUNG_obsgermany_climate_10min_precipitation_de.pdf
    "RWS_10": {"core_name": "LiquidPrecD", "time_of_meas_shift": "prec2ind", "unit": "mm/h", "resample": "sum"},
    # https://opendata.dwd.de/climate_environment/CDC/observations_germany/climate/hourly/soil_temperature/BESCHREIBUNG_obsgermany_climate_hourly_soil_temperature_de.pdf
    "V_TE100": {"core_name": "Soil_Temperature_1m", "time_of_meas_shift": None, "unit": "degC"},
    "V_TE050": {"core_name": "Soil_Temperature_50cm", "time_of_meas_shift": None, "unit": "degC"},
    "V_TE020": {"core_name": "Soil_Temperature_20cm", "time_of_meas_shift": None, "unit": "degC"},
    "V_TE010": {"core_name": "Soil_Temperature_10cm", "time_of_meas_shift": None, "unit": "degC"},
    "V_TE005": {"core_name": "Soil_Temperature_5cm", "time_of_meas_shift": None, "unit": "degC"},
    # https://opendata.dwd.de/climate_environment/CDC/observations_germany/climate/hourly/cloud_type/BESCHREIBUNG_obsgermany_climate_hourly_cloud_type_de.pdf
    " V_N": {"core_name": "TotalSkyCover", "time_of_meas_shift": None, "unit": "1eighth"},
    # Hourly measurements currently unused due to doubling with 10
    # minute data and conflicting time shifting and units
    # "RF_TU": "RelHum",
    # "TT_TU": "DryBulbTemp",
    # "  P0": "AtmPressure",
    # "   P": "Pressure_Sea_Level",
    # "   F": "WindSpeed",
    # "   D": "WindDir",
    # "  R1": "LiquidPrecD",
}
# get variables that should be resampled by sum instead of mean
variables_to_sum_DWD_historical = [
    key for key, value in format_DWD_historical.items() if "resample" in value.keys()
]

"""
format_DWD_forecast information:

Variable definitions: https://opendata.dwd.de/weather/lib/MetElementDefinition.xml or 
https://wetterdienst.readthedocs.io/en/latest/data/coverage/dwd/mosmix/hourly.html (in origin unit)

checked by Martin Rätz 18.08.2023
"""
format_DWD_forecast = {
    # "cloud_cover_above_7_km": None,
    # "cloud_cover_below_1000_ft": None,
    # "cloud_cover_below_500_ft": None,
    # "cloud_cover_between_2_to_7_km": None,
    "cloud_cover_effective": {"core_name": "OpaqueSkyCover", "time_of_meas_shift": None, "unit": "%"},
    "cloud_cover_total": {"core_name": "TotalSkyCover", "time_of_meas_shift": None, "unit": "%"},
    # "precipitation_height_significant_weather_last_1h": None,
    # "precipitation_height_significant_weather_last_3h": None,
    "pressure_air_site_reduced": {"core_name": "AtmPressure", "time_of_meas_shift": None, "unit": "Pa"},
    # "probability_fog_last_12h": None,
    # "probability_fog_last_1h": None,
    # "probability_fog_last_6h": None,
    # "probability_precipitation_height_gt_0_0_mm_last_12h": None,
    # "probability_precipitation_height_gt_0_2_mm_last_12h": None,
    # "probability_precipitation_height_gt_0_2_mm_last_24h": None,
    # "probability_precipitation_height_gt_0_2_mm_last_6h": None,
    # "probability_precipitation_height_gt_1_0_mm_last_12h": None,
    # "probability_precipitation_height_gt_5_0_mm_last_12h": None,
    # "probability_precipitation_height_gt_5_0_mm_last_24h": None,
    # "probability_precipitation_height_gt_5_0_mm_last_6h": None,
    # "probability_wind_gust_ge_25_kn_last_12h": None,
    # "probability_wind_gust_ge_40_kn_last_12h": None,
    # "probability_wind_gust_ge_55_kn_last_12h": None,
    # is actually balance during the last 3 hours:
    "radiation_global": {"core_name": "GlobHorRad", "time_of_meas_shift": "prec2ind", "unit": "kJ/m2"},
    # "sunshine_duration": None,
    # "temperature_air_max_200": None,
    # "temperature_air_mean_005": None,
    # no information if temperature is drybulb or something else:
    "temperature_air_mean_200": {"core_name": "DryBulbTemp", "time_of_meas_shift": None, "unit": "K"},
    # "temperature_air_min_200": None,
    "temperature_dew_point_mean_200": {"core_name": "DewPointTemp", "time_of_meas_shift": None, "unit": "K"},
    "visibility_range": {"core_name": "Visibility", "time_of_meas_shift": None, "unit": "m"},
    # "water_equivalent_snow_depth_new_last_1h": None,
    # "water_equivalent_snow_depth_new_last_3h": None,
    # "weather_last_6h": None,
    # "weather_significant": None,
    "wind_direction": {"core_name": "WindDir", "time_of_meas_shift": None, "unit": "deg"},
    # "wind_gust_max_last_12h": None,
    # "wind_gust_max_last_1h": None,
    # "wind_gust_max_last_3h": None,
    "wind_speed": {"core_name": "WindSpeed", "time_of_meas_shift": None, "unit": "m/s"}
}


def DWD_historical_to_core_data(
    df_import: pd.DataFrame, start: datetime, stop: datetime, meta: MetaData
) -> pd.DataFrame:
    """
    Transform imported weather data from DWD historical format into core data format.

    Args:
        df_import (pd.DataFrame): The DataFrame containing imported weather data from DWD.
        start (datetime): The timestamp for the start of the desired data range (will be extended for interpolation).
        stop (datetime): The timestamp for the end of the desired data range (will be extended for interpolation).
        meta (MetaData): Metadata associated with the data.

    Returns:
        pd.DataFrame: The transformed DataFrame in the core data format.
    """

    ### evaluate correctness of format
    auxiliary.evaluate_transformations(
        core_format=definitions.format_core_data, other_format=format_DWD_historical
    )

    ### format raw data for further operations
    df = df_import.copy()
    # to datetime; account for different time-formats
    date_format = "%Y%m%d%H%M"
    df.index = pd.to_datetime(df.index, format=date_format)
    # sort by time
    df = df.sort_index()

    # reduce time period to extended period for working interpolation and for faster operation
    df = time_observation_transformations.truncate_data_from_start_to_stop(
        df, start - timedelta(days=1), stop + timedelta(days=1)
    )

    # select only numeric columns
    df = df.select_dtypes(include=["number"])

    # delete dummy values from DWD
    df = auxiliary.replace_dummy_with_nan(df, format_DWD_historical)

    # resample some via sum some via mean -> results in average of following hour
    df = df.resample("H").mean()  # fills nan only if all nan in interval
    for var in variables_to_sum_DWD_historical:
        if var in df.columns:
            df_sum = (
                df[var].resample("H").sum(min_count=1)
            )  # fills nan only if 1 value in interval
            df[var] = df_sum

    # rename available variables to core data format
    df = auxiliary.rename_columns(df, format_DWD_historical)

    ### convert timezone to UTC
    # the data is for most stations and datasets, as well as for more recent
    # data (several years) in UTC. For more sophisticated handling pull meta
    # and respect time zone or implement dwd_pulling repo from github

    ### shift and interpolate data forward 30mins or backward -30mins
    df_no_shift = df.copy()
    df = time_observation_transformations.shift_time_by_dict(format_DWD_historical, df)

    def transform_DWD_historical(df):
        # drop unnecessary variables
        df = auxiliary.force_data_variable_convention(df, definitions.format_core_data)

        ### convert units
        df["AtmPressure"] = unit_conversions.hPa_to_Pa(df["AtmPressure"])
        df["DiffHorRad"] = unit_conversions.Jcm2_to_Whm2(df["DiffHorRad"])
        df["GlobHorRad"] = unit_conversions.Jcm2_to_Whm2(df["GlobHorRad"])
        df["HorInfra"] = unit_conversions.Jcm2_to_Whm2(df["HorInfra"])
        df["TotalSkyCover"] = unit_conversions.eigth_to_tenth(df["TotalSkyCover"])

        ### impute missing variables from other available ones
        df, calc_overview = variable_transformations.variable_transform_all(df, meta)

        return df, calc_overview

    df, meta.executed_transformations = transform_DWD_historical(df)

    ### add unshifted data for possible later direct use (pass-through),
    ### to avoid back and forth interpolating
    df = pass_through_handling.create_pass_through_variables(
        df_shifted=df,
        df_no_shift=df_no_shift,
        format=format_DWD_historical,
        transform_func=transform_DWD_historical,
        meta=meta,
    )

    return df


def DWD_forecast_2_core_data(df_import: pd.DataFrame, meta: MetaData) -> pd.DataFrame:
    """
    Transform imported weather forecast data from DWD into core data format.

    Args:
        df_import (pd.DataFrame): The DataFrame containing imported weather forecast data from DWD.
        meta (MetaData): Metadata associated with the data.

    Returns:
        pd.DataFrame: The transformed DataFrame in the core data format.
    """

    ### evaluate correctness of format
    auxiliary.evaluate_transformations(
        core_format=definitions.format_core_data, other_format=format_DWD_forecast
    )

    ### format raw data for further operations
    df = df_import.copy()
    # Resample the DataFrame to make the DatetimeIndex complete and monotonic
    df = df.resample('H').asfreq()
    # delete timezone information
    df = df.tz_localize(None)
    # rename available variables to core data format
    df = auxiliary.rename_columns(df, format_DWD_forecast)

    ### convert timezone to UTC
    # the data pulled by Wetterdienst is already UTC

    ### shift and interpolate data forward 30mins or backward -30mins
    df_no_shift = df.copy()
    df = time_observation_transformations.shift_time_by_dict(format_DWD_forecast, df)

    def transform_DWD_forecast(df):
        # drop unnecessary variables
        df = auxiliary.force_data_variable_convention(df, definitions.format_core_data)

        ### convert units
        df["OpaqueSkyCover"] = unit_conversions.percent_to_tenth(df["OpaqueSkyCover"])
        df["TotalSkyCover"] = unit_conversions.percent_to_tenth(df["TotalSkyCover"])
        df["GlobHorRad"] = unit_conversions.kJm2_to_Whm2(df["GlobHorRad"])
        df["DryBulbTemp"] = unit_conversions.kelvin_to_celcius(df["DryBulbTemp"])
        df["DewPointTemp"] = unit_conversions.kelvin_to_celcius(df["DewPointTemp"])
        df["Visibility"] = unit_conversions.divide_by_1000(df["Visibility"])

        ### impute missing variables from other available ones
        df, calc_overview = variable_transformations.variable_transform_all(df, meta)

        return df, calc_overview

    df, meta.executed_transformations = transform_DWD_forecast(df)

    ### add unshifted data for possible later direct use (pass-through),
    ### to avoid back and forth interpolating
    df = pass_through_handling.create_pass_through_variables(
        df_shifted=df,
        df_no_shift=df_no_shift,
        format=format_DWD_forecast,
        transform_func=transform_DWD_forecast,
        meta=meta,
    )

    return df



"""
Converts core data to modelica TMY3Reader data
"""

import calendar
import datetime as dt
import pandas as pd

from aixweather import definitions
from aixweather.imports.utils_import import MetaData
from aixweather.transformation_functions import auxiliary, time_observation_transformations, pass_through_handling

"""
format_modelica_TMY3 information:

Please indicate the unit (conversion has to be changed manually)
Please indicate the desired time of measurement (conversion automatically)
The stated nans are the default values stated from the AixLib TMY3 Reader, those values are filled if nan.

time_of_meas_shift´s checked by Martin Rätz (07.08.2023)
unit´s checked by Martin Rätz (07.08.2023)
"""
format_modelica_TMY3 = {
    'timeOfYear': {'core_name': '', 'unit': 's', 'time_of_meas_shift': None, 'nan': None},
    'DryBulbTemp': {'core_name': 'DryBulbTemp', 'unit': 'degC', 'time_of_meas_shift': None, 'nan': 20.0},
    'DewPointTemp': {'core_name': 'DewPointTemp', 'unit': 'degC', 'time_of_meas_shift': None, 'nan': 10.0},
    'RelHum': {'core_name': 'RelHum', 'unit': 'percent', 'time_of_meas_shift': None, 'nan': 50},
    'AtmPressure': {'core_name': 'AtmPressure', 'unit': 'Pa', 'time_of_meas_shift': None, 'nan': 101325},
    'ExtHorRad': {'core_name': 'ExtHorRad', 'unit': 'Wh/m2', 'time_of_meas_shift': 'ind2prec', 'nan': '-0'},
    'ExtDirNormRad': {'core_name': 'ExtDirNormRad', 'unit': 'Wh/m2', 'time_of_meas_shift': 'ind2prec', 'nan': '-0'},
    'HorInfra': {'core_name': 'HorInfra', 'unit': 'Wh/m2', 'time_of_meas_shift': 'ind2prec', 'nan': 0},
    'GlobHorRad': {'core_name': 'GlobHorRad', 'unit': 'Wh/m2', 'time_of_meas_shift': 'ind2prec', 'nan': 0},
    'DirNormRad': {'core_name': 'DirNormRad', 'unit': 'Wh/m2', 'time_of_meas_shift': 'ind2prec', 'nan': 0},
    'DiffHorRad': {'core_name': 'DiffHorRad', 'unit': 'Wh/m2', 'time_of_meas_shift': 'ind2prec', 'nan': 0},
    'GlobHorIll': {'core_name': 'GlobHorIll', 'unit': 'lux', 'time_of_meas_shift': 'ind2prec', 'nan': '-0'},
    'DirecNormIll': {'core_name': 'DirecNormIll', 'unit': 'lux', 'time_of_meas_shift': 'ind2prec', 'nan': '-0'},
    'DiffuseHorIll': {'core_name': 'DiffuseHorIll', 'unit': 'lux', 'time_of_meas_shift': 'ind2prec', 'nan': '-0'},
    'ZenithLum': {'core_name': 'ZenithLum', 'unit': 'Cd/m2', 'time_of_meas_shift': 'ind2prec', 'nan': '-0'},
    'WindDir': {'core_name': 'WindDir', 'unit': 'deg', 'time_of_meas_shift': None, 'nan': '-0'},
    'WindSpeed': {'core_name': 'WindSpeed', 'unit': 'm/s', 'time_of_meas_shift': None, 'nan': '-0'},
    'TotalSkyCover': {'core_name': 'TotalSkyCover', 'unit': '1tenth', 'time_of_meas_shift': None, 'nan': 5},
    'OpaqueSkyCover': {'core_name': 'OpaqueSkyCover', 'unit': '1tenth', 'time_of_meas_shift': None, 'nan': 5},
    'Visibility': {'core_name': 'Visibility', 'unit': 'km', 'time_of_meas_shift': None, 'nan': '-0'},
    'CeilingH': {'core_name': 'CeilingH', 'unit': 'm', 'time_of_meas_shift': None, 'nan': 20000.0},
    'WeatherObs': {'core_name': '', 'unit': '', 'time_of_meas_shift': None, 'nan': '-0'},
    'WeatherCode': {'core_name': '', 'unit': '', 'time_of_meas_shift': None, 'nan': '-0'},
    'PrecWater': {'core_name': 'PrecWater', 'unit': 'mm', 'time_of_meas_shift': None, 'nan': '-0'},
    'Aerosol': {'core_name': 'Aerosol', 'unit': '1thousandth', 'time_of_meas_shift': None, 'nan': '-0'},
    'Snow': {'core_name': '', 'unit': 'cm', 'time_of_meas_shift': None, 'nan': '-0'},
    'DaysSinceSnow': {'core_name': '', 'unit': 'days', 'time_of_meas_shift': None, 'nan': '-0'},
    'Albedo': {'core_name': '', 'unit': '', 'time_of_meas_shift': None, 'nan': '-0'},
    'LiquidPrecD': {'core_name': 'LiquidPrecD', 'unit': 'mm/h', 'time_of_meas_shift': None, 'nan': '-0'},
    'LiquidPrepQuant': {'core_name': '', 'unit': '', 'time_of_meas_shift': None, 'nan': '-0'}
}


def to_mos(
    core_df: pd.DataFrame,
    meta: MetaData,
    start: dt.datetime,
    stop: dt.datetime,
    fillna: bool,
    result_folder: str = None
) -> pd.DataFrame:
    """Create a MOS file from the core data.

    Args:
        core_df (pd.DataFrame): DataFrame containing core data.
        meta (MetaData): Metadata associated with the weather data.
        start (dt.datetime): Timestamp for the start of the MOS file.
        stop (dt.datetime): Timestamp for the end of the MOS file.
        fillna (bool): Boolean indicating whether NaN values should be filled.

    Returns:
        pd.DataFrame: DataFrame containing the weather data formatted for MOS export,
                      excluding metadata.
    """
    ### evaluate correctness of format
    auxiliary.evaluate_transformations(
        core_format=definitions.format_core_data, other_format=format_modelica_TMY3
    )

    df = core_df.copy()

    ### measurement time conversion
    df = time_observation_transformations.shift_time_by_dict(format_modelica_TMY3, df)

    ### if possible avoid back and forth interpolating -> pass through variables without shifting
    df = pass_through_handling.pass_through_measurements_with_back_and_forth_interpolating(
        df, format_modelica_TMY3
    )

    ### select only desired period
    df = time_observation_transformations.truncate_data_from_start_to_stop(
        df, start, stop
    )

    ### select the desired columns
    df = auxiliary.force_data_variable_convention(df, format_modelica_TMY3)

    # tmy3 data must start with 0 at the beginning of year (due to internal
    # tmy3_reader sun inclination calculation)
    min_year = min(df.index.year)
    time_of_year = (
            (df.index.year - min_year) * 365 * 24 * 3600
            + calendar.leapdays(min_year, df.index.year) * 24 * 3600
            + (df.index.dayofyear - 1) * 24 * 3600
            + df.index.hour * 3600
    )
    df["timeOfYear"] = time_of_year

    # to avoid having the one-year duration between start and end of data as
    # it will make the tmy3_reader loop the data
    if (df["timeOfYear"].iloc[-1] - df["timeOfYear"].iloc[0]) == 365 * 24 * 3600:
        # copy last row
        last_row = df.iloc[-1]

        # Convert the Series to a new DataFrame with a single row
        new_row_df = last_row.to_frame().T

        # continue time values
        new_row_df["timeOfYear"].iloc[0] = new_row_df["timeOfYear"].iloc[0] + 3600
        new_row_df.index = new_row_df.index + dt.timedelta(hours=1)

        # add new row to df, make sure its int
        df = pd.concat([df, new_row_df])
        df["timeOfYear"] = df["timeOfYear"].astype(int)

    ### fill nan
    if fillna:
        # fill first and last lines nans (possibly lost through shifting)
        df.iloc[0, :] = df.bfill().iloc[0, :]
        df.iloc[-1, :] = df.ffill().iloc[-1, :]
        # fill dummy values
        auxiliary.fill_nan_from_format_dict(df, format_modelica_TMY3)

    ### Create header
    header_of = (
        "#1:for TMY3reader"
        + "\ndouble tab1("
        + str(int(df.index.size))
        + ","
        + str(int(df.columns.size))
        + ")"
    )
    header_of += f"\n#LOCATION,{meta.station_name},,,,,{str(meta.latitude)}," \
                 f"{str(meta.longitude)},0,something"
    header_of += (
        "\n#Explanation of Location line:"
        + "\n#   Element 7: latitude"
        + "\n#   Element 8: longitude"
        + "\n#   Element 9: time zone in hours from UTC"
        + "\n#"
    )

    # Data periods
    header_of += (
        "\n#DATA PERIODS, data available from "
        + str(dt.datetime.strptime(str(df.index[0]), "%Y-%m-%d %H:%M:%S"))
        + " "
        + "(second="
        + str(df["timeOfYear"].iloc[0])
        + ") to "
        + str(dt.datetime.strptime(str(df.index[-1]), "%Y-%m-%d %H:%M:%S"))
        + " "
        + "(second="
        + str(df["timeOfYear"].iloc[-1])
        + ")"
        + "\n# info: TMY3Reader requirement: Time 0 = 01.01. 00:00:00 at local"
          " time (see time zone above)"
        + "\n#"
    )
    # data source
    header_of += (
        "\n#USED DATA-COLLECTOR: "
        + "ebc-weather-tool with input source "
        + str(meta.input_source)
        + " (collected at "
        + str(dt.datetime.now())
        + ') "-0" marks not available data'
        "\n#Info: The AixLib/IBPSA TMY3reader requires the below mentioned units and"
          " measurement times. Last Check: 28.02.2022"
    )

    # Information about the data
    header_of += (
        "\n#C1 Time in seconds. Beginning of a year is 0s."
        + "\n#C2 Dry bulb temperature in Celsius at indicated time"
        + "\n#C3 Dew point temperature in Celsius at indicated time"
        + "\n#C4 Relative humidity in percent at indicated time"
        + "\n#C5 Atmospheric station pressure in Pa at indicated time, TMY3Reader:"
          " not used per default"
        + "\n#C6 Extraterrestrial horizontal radiation in Wh/m2, TMY3Reader: not used"
        + "\n#C7 Extraterrestrial direct normal radiation in Wh/m2, TMY3Reader: not used"
        + "\n#C8 Horizontal infrared radiation intensity in Wh/m2"
        + "\n#C9 Global horizontal radiation in Wh/m2"
        + "\n#C10 Direct normal radiation in Wh/m2"
        + "\n#C11 Diffuse horizontal radiation in Wh/m2"
        + "\n#C12 Averaged global horizontal illuminance in lux during minutes preceding"
          " the indicated time, TMY3Reader: not used"
        + "\n#C13 Direct normal illuminance in lux during minutes preceding the"
          " indicated time, TMY3Reader: not used"
        + "\n#C14 Diffuse horizontal illuminance in lux during minutes preceding the"
          " indicated time, TMY3Reader: not used"
        + "\n#C15 Zenith luminance in Cd/m2 during minutes preceding the indicated"
          " time, TMY3Reader: not used"
        + "\n#C16 Wind direction at indicated time. N=0, E=90, S=180, W=270"
        + "\n#C17 Wind speed in m/s at indicated time"
        + "\n#C18 Total sky cover in tenth at indicated time"
        + "\n#C19 Opaque sky cover tenth indicated time"
        + "\n#C20 Visibility in km at indicated time, TMY3Reader: not used"
        + "\n#C21 Ceiling height in m"
        + "\n#C22 Present weather observation, TMY3Reader: not used"
        + "\n#C23 Present weather codes, TMY3Reader: not used"
        + "\n#C24 Precipitable water in mm, TMY3Reader: not used"
        + "\n#C25 Aerosol optical depth, TMY3Reader: not used"
        + "\n#C26 Snow depth in cm, TMY3Reader: not used"
        + "\n#C27 Days since last snowfall, TMY3Reader: not used"
        + "\n#C28 Albedo, TMY3Reader: not used"
        + "\n#C29 Liquid precipitation depth in mm/h at indicated time, TMY3Reader: not used"
        + "\n#C30 Liquid precipitation quantity, TMY3Reader: not used"
    )

    ### write to csv
    filename = (
        f"{meta.station_id}_{start.strftime('%Y%m%d')}_{stop.strftime('%Y%m%d')}"
        f"_{meta.station_name}.mos"
    )
    filepath = definitions.results_file_path(filename, result_folder)

    df.to_csv(
        filepath,
        sep="\t",
        float_format="%.2f",
        header=False,
        index_label="timeOfYear",
        index=False,
    )

    # Read the contents and prepend the header_of to the file
    with open(filepath, "r+") as file:
        content = file.read()
        file.seek(0, 0)
        file.write(f"{header_of}\n{content}")

    print(f"MOS file saved to {filepath}.")
    return df



"""This module contains the central project classes which are used by the user."""

from abc import ABC, abstractmethod
import datetime as dt
import pandas as pd

from aixweather.imports.DWD import (
    import_DWD_historical,
    import_DWD_forecast,
    import_meta_DWD_historical,
    import_meta_DWD_forecast,
)
from aixweather.imports.ERC import import_ERC, import_meta_from_ERC
from aixweather.imports.TRY import load_try_from_file, load_try_meta_from_file
from aixweather.imports.EPW import load_epw_from_file, load_epw_meta_from_file
from aixweather.imports.custom_file import load_custom_meta_data, load_custom_from_file
from aixweather.transformation_to_core_data.DWD import (
    DWD_historical_to_core_data,
    DWD_forecast_2_core_data,
)
from aixweather.transformation_to_core_data.ERC import ERC_to_core_data
from aixweather.transformation_to_core_data.TRY import TRY_to_core_data
from aixweather.transformation_to_core_data.EPW import EPW_to_core_data
from aixweather.transformation_to_core_data.custom_file import custom_to_core_data
from aixweather.core_data_format_2_output_file.unconverted_to_x import (
    to_pickle,
    to_json,
    to_csv,
)
from aixweather.core_data_format_2_output_file.to_mos_TMY3 import to_mos
from aixweather.core_data_format_2_output_file.to_epw_energyplus import to_epw


# pylint-disable: R0902
class ProjectClassGeneral(ABC):
    """
    An abstract base class representing a general project.

    For each source of weather data, a project class should inherit from this class
    and implement specific methods for data import and transformation.

    Attributes:
        fillna (bool): A flag indicating whether NaN values should be filled
            in the output formats.
        abs_result_folder_path (str): Optionally define the absolute path to
            the desired export location.
        start (pd.Timestamp or None): The start date of the project data
            (sometimes inferred by the inheriting class).
        end (pd.Timestamp or None): The end date of the project data.

    Properties:
        imported_data (pd.DataFrame): The imported weather data.
        core_data (pd.DataFrame): The weather data in a standardized core
            format.
        output_df_<outputformat> (pd.DataFrame): The output data frame
            (name depends on output format).
        meta_data: Metadata associated with weather data origin.

    Methods:
        import_data(): An abstract method to import data from the specific
            source.
        data_2_core_data(): An abstract method to transform imported data into
            core data format.
        core_2_mos(): Convert core data to MOS format.
        core_2_epw(): Convert core data to EPW format.
        core_2_csv(): Convert core data to CSV format.
        core_2_json(): Convert core data to JSON format.
        core_2_pickle(): Convert core data to Pickle format.
    """

    def __init__(self, **kwargs):
        # User-settable attributes
        self.fillna = kwargs.get(
            "fillna", True
        )  # defines whether nan should be filled in the output formats
        self.abs_result_folder_path = kwargs.get("abs_result_folder_path", None)

        # User-settable or placeholder depending on data origin
        self.start = kwargs.get("start", None)
        self.end = kwargs.get("end", None)

        # Placeholder attributes (set during execution)
        self.imported_data: pd.DataFrame = None
        self.core_data: pd.DataFrame = None
        self.output_data_df: pd.DataFrame = None
        self.meta_data = None

    @property
    def imported_data(self):
        """Get imported data"""
        return self._imported_data

    @imported_data.setter
    def imported_data(self, data):
        """If the imported data is empty, the program should be stopped with clear error
        description."""
        if isinstance(data, pd.DataFrame) and data.empty:
            raise ValueError(
                "Imported data cannot be an empty DataFrame. No weather data "
                "has been found. Possible reasons are: "
                "The station id is wrong or the station does not exist/measure "
                "anymore, the data is currently not available or the weather "
                "provider can not be accessed.\n"
                f"The so far pulled meta data is: {self.meta_data.__dict__}"
            )
        self._imported_data = data

    @property
    def core_data(self):
        """Get core data"""
        return self._core_data

    @core_data.setter
    def core_data(self, value: pd.DataFrame):
        """Makes sure the core-data data types are correct."""
        if value is not None:
            for column in value.columns:
                # only real pd.NA values
                # force strings to be NaN
                value[column] = pd.to_numeric(value[column], errors="coerce")
                # round floats for unit test compatibility across different machines
                digits_2_round = 5
                if value[column].dtype == "float":
                    value[column] = value[column].round(digits_2_round)

            self._core_data = value

    @abstractmethod
    def import_data(self):
        """Abstract function to import weather data."""

    @abstractmethod
    def data_2_core_data(self):
        """Abstract function to convert the imported data to core data."""

    # core_data_format_2_output_file
    def core_2_mos(self):
        """Convert core data to .mos file"""
        self.output_data_df = to_mos(
            self.core_data,
            self.meta_data,
            self.start,
            self.end,
            self.fillna,
            self.abs_result_folder_path,
        )

    def core_2_epw(self):
        """Convert core data to .epw file"""
        self.output_data_df = to_epw(
            self.core_data,
            self.meta_data,
            self.start,
            self.end,
            self.fillna,
            self.abs_result_folder_path,
        )

    def core_2_csv(self):
        """Convert core data to .csv file"""
        self.output_data_df = to_csv(
            self.core_data, self.meta_data, self.abs_result_folder_path
        )

    def core_2_json(self):
        """Convert core data to .json file"""
        self.output_data_df = to_json(
            self.core_data, self.meta_data, self.abs_result_folder_path
        )

    def core_2_pickle(self):
        """Convert core data pickle file"""
        self.output_data_df = to_pickle(
            self.core_data, self.meta_data, self.abs_result_folder_path
        )


class ProjectClassDWDHistorical(ProjectClassGeneral):
    """
    A class representing a project for importing and processing historical weather data
    from DWD (Deutscher Wetterdienst).

    For common attributes, properties, and methods, refer to the base class.

    Attributes:
        station (str): The identifier of the DWD weather station associated with the data.
    """

    def __init__(self, start: dt.datetime, end: dt.datetime, station: str, **kwargs):
        super().__init__(start=start, end=end, **kwargs)
        self.station = str(station)

    # imports
    def import_data(self):
        """override abstract function"""
        self.meta_data = import_meta_DWD_historical(station=self.station)
        self.imported_data = import_DWD_historical(
            self.start - dt.timedelta(days=1), self.station
        )  # pull more data for better interpolation

    # transformation_2_core_data_DWD_Historical
    def data_2_core_data(self):
        """override abstract function"""
        self.core_data = DWD_historical_to_core_data(
            self.imported_data,
            self.start - dt.timedelta(days=1),
            self.end + dt.timedelta(days=1),
            self.meta_data,
        )


class ProjectClassDWDForecast(ProjectClassGeneral):
    """
    A class representing a project for importing and processing weather forecast data
    from DWD (Deutscher Wetterdienst).

    For common attributes, properties, and methods, refer to the base class.

    Attributes:
        station (str): The identifier of the KML grid associated with the forecast data.
    """

    def __init__(self, station: str, **kwargs):
        super().__init__(**kwargs)
        self.station = str(station)

    # imports
    def import_data(self):
        """override abstract function"""
        self.meta_data = import_meta_DWD_forecast(self.station)
        self.imported_data = import_DWD_forecast(self.station)

    def data_2_core_data(self):
        """override abstract function"""
        self.core_data = DWD_forecast_2_core_data(self.imported_data, self.meta_data)
        self.start = self.core_data.index[0]
        self.end = self.core_data.index[-1]


class ProjectClassERC(ProjectClassGeneral):
    """
    A class representing a project for importing and processing weather data
    from the ERC (Energy Research Center).

    For common attributes, properties, and methods, refer to the base class.

    Attributes:
        cred (tuple): A tuple containing credentials or authentication information for accessing
        the data source.
    """

    def __init__(
        self, start: dt.datetime, end: dt.datetime, cred: tuple = None, **kwargs
    ):
        super().__init__(start=start, end=end, **kwargs)
        self.cred = cred
        self.start_hour_earlier = start - dt.timedelta(hours=2)
        self.end_hour_later = end + dt.timedelta(hours=2)

    def import_data(self):
        """override abstract function"""
        self.meta_data = import_meta_from_ERC()
        self.imported_data = import_ERC(
            self.start_hour_earlier, self.end_hour_later, self.cred
        )

    def data_2_core_data(self):
        """override abstract function"""
        self.core_data = ERC_to_core_data(self.imported_data, self.meta_data)


class ProjectClassTRY(ProjectClassGeneral):
    """
    A class representing a project for importing and processing weather data
    from TRY (Test Reference Year) format.

    For common attributes, properties, and methods, refer to the base class.

    Attributes:
        path (str): The absolute file path to the TRY weather data.
    """

    def __init__(self, path, **kwargs):
        super().__init__(**kwargs)
        self.path = path

    # imports
    def import_data(self):
        """override abstract function"""
        self.meta_data = load_try_meta_from_file(path=self.path)
        self.imported_data = load_try_from_file(path=self.path)

    # transformation_2_core_data_TRY
    def data_2_core_data(self):
        """override abstract function"""
        self.core_data = TRY_to_core_data(self.imported_data, self.meta_data)
        self.start = self.core_data.index[0]
        self.end = self.core_data.index[-1]


class ProjectClassEPW(ProjectClassGeneral):
    """
    A class representing a project for importing and processing weather data
    from EPW (EnergyPlus Weather) format.

    For common attributes, properties, and methods, refer to the base class.

    Attributes:
        path (str): The absolute file path to the EPW weather data.
    """

    def __init__(self, path, **kwargs):
        super().__init__(**kwargs)
        self.path = path

    # imports
    def import_data(self):
        """override abstract function"""
        self.meta_data = load_epw_meta_from_file(path=self.path)
        self.imported_data = load_epw_from_file(path=self.path)

    # transformation_2_core_data_TRY
    def data_2_core_data(self):
        """override abstract function"""
        self.core_data = EPW_to_core_data(self.imported_data, self.meta_data)
        self.start = self.core_data.index[0]
        self.end = self.core_data.index[-1]


class ProjectClassCustom(ProjectClassGeneral):
    """
    A class representing a project for importing and processing custom weather data.
    Modify this class and its functions to create your own weather data pipeline
    and consider to create a pull request to add the pipeline to the repository.

    For common attributes, properties, and methods, refer to the base class.

    Attributes:
        path (str): The file path to the custom weather data.
    """

    def __init__(self, path, **kwargs):
        super().__init__(**kwargs)
        self.path = path

    # imports
    def import_data(self):
        """override abstract function"""
        self.meta_data = load_custom_meta_data()
        self.imported_data = load_custom_from_file(path=self.path)

    # transformation_2_core_data_TRY
    def data_2_core_data(self):
        """override abstract function"""
        self.core_data = custom_to_core_data(self.imported_data, self.meta_data)
        self.start = self.core_data.index[0]  # or define in init
        self.end = self.core_data.index[-1]  # or define in init
